#include "main.h"
#include "ps2_kbd.h"
#include "Keyboard.h"
#include "ps2_scancodes.h"

// state
bool got_keyup = false;		// true if received key-released byte (0xF0)
bool got_2byte = false;		// true if receiving two-byte, four-byte, or eight-byte scancode
bool got_4byte = false;
bool got_8byte = false;
uint8_t byte_ctr;			// when multi-byte scancode indicator received, set to number of bytes remaining in scancode

void kbd_init (void) {
	// clock and data pins are inputs by default,
	// which is how they should BE

	//TODO: wait for clock line to go high before enabling receiver?

	// setup USART
	UCSR1C &= ~B(UMSEL11);	// synchronous USART
	UCSR1C |= B(UMSEL10);
	UCSR1C |= B(UPM11);		// odd parity
	UCSR1C |= B(UPM10);
	UCSR1C &= ~B(USBS1);	// one stop bit
	UCSR1B &= ~B(UCSZ12);	// 8 data bits
	UCSR1C |= B(UCSZ11);
	UCSR1C |= B(UCSZ10);
	UCSR1C |= B(UCPOL1);	// data sampled on rising edge of clk, set on falling edge of clk
	UCSR1B |= B(RXEN1);		// enable receiver
//	UCSR1B |= B(TXEN1);		// enable transmitter
}

void kbd_set_leds (uint8_t leds) {
	// don't do anything if we're in the middle of receiving a scancode
	if (got_keyup || got_2byte || got_4byte || got_8byte)
		return;

	// TODO: debug - for some reason seems to be trying to start transmitting in the middle of
	// the keyboard sending data
	// no ways to see if receiver is busy...why's it trying to transmit always in the middle of a reception though?

	// send "set LEDs" command code first
	UDR1 = 0xED;
	// wait for that to finish
	while (UCSR1A & !B(TXC1));
	UCSR1A = B(TXC1);
	// next send the LED status byte
	UDR1 = leds;
	// wait for that to finish too
	while (UCSR1A & !B(TXC1));
	UCSR1A = B(TXC1);
}

bool kbd_poll (uint8_t *hid_scancode, bool *modifier, bool *pressed) {
	uint8_t at_scancode;	// received scancode byte
	uint8_t *table;			// AT-to-HID scancode table to use

	// check if there's a new byte received
	if (UCSR1A & B(RXC1)) {
		// if so, make sure it's valid (no parity errors, framing errors, etc.)
		if (UCSR1A & (B(FE1) | B(DOR1) | B(UPE1))) {
			// if errors, read the invalid byte to remove it from the buffer,
			// and ignore it
			*hid_scancode = UDR1;
			return false;
		}

		// the byte's valid, so read it
		at_scancode = UDR1;
		// there are a few options here:
		switch (at_scancode) {
			case 0xF0:
				if (got_8byte) {
					// 0xF0 occurs in the middle of the 8-byte scancode
					// check that it's in the right position
					if ((byte_ctr == 4) || (byte_ctr == 2)) {
						byte_ctr--;
						return false;
					} else goto error;
				} else {
					// 0xF0 at beginning means that a key has been released, and the key's scancode will follow
					got_keyup = true;
					return false;
				}
				break;

			case 0xE0:
				if (got_4byte) {
					// 0xE0 occurs in the middle of the four-byte scancode
					// check that it's in the right position
					if (byte_ctr == 2) {
						byte_ctr--;
						return false;
					} else goto error;
				} else {
					// 0xE0 at beginning means that a two-byte or four-byte scancode will follow
					// until we know it's a four-byte scancode, assume two-byte
					got_2byte = true;
					return false;
				}
				break;

			case 0xE1:
				if (got_8byte) {
					// 0xE1 occurs in the middle of the 8-byte scancode
					// check that it's in the right position
					if (byte_ctr == 5) {
						byte_ctr--;
						return false;
					} else goto error;
				} else {
					// 0xE1 at beginning means that an eight-byte scancode will follow (only used for one key - Break)
					got_8byte = true;
					byte_ctr = 7;
					return false;
				}
				break;

			default:
				if (got_2byte) {
					// if it's a two-byte scancode...
					if (at_scancode == 0x12) {
						// ...and the second byte is 0x12, that means it's actually a four-byte scancode
						got_2byte = false;
						got_4byte = true;
						byte_ctr = 2;
						return false;
					}

					// otherwise, the two-byte scancode is done, so process it
					table = at_2byte;

				} else if (got_4byte) {
					// if it's a four-byte scancode, make sure it's the correct value (since only one key matches this)
					if ((at_scancode == 0x7C) && (byte_ctr == 1)) {
						table = NULL;
						*modifier = false;
						*hid_scancode = HID_KEYBOARD_SC_PRINT_SCREEN;
					} else goto error;	// if incorrect, there's been an error - reset status and ignore

				} else if (got_8byte) {
					switch (byte_ctr) {
						case 7:
						case 3:
							if (at_scancode == 0x14) return false; else goto error; break;
						case 6:
							if (at_scancode == 0x77) return false; else goto error; break;
						case 1:
							if (at_scancode == 0x77) {
								table = NULL;
								*modifier = false;
								*hid_scancode = HID_KEYBOARD_SC_PAUSE;
							} else goto error;
							break;
						default:
							goto error;
					}
					byte_ctr--;

				} else {
					// it's just a normal one-byte scancode
					table = at_1byte;
				}

				break;
		}

		// if table is NULL, then the HID scancode has already been set
		// otherwise, we need to look it up in the table from at_scancode
		if (table != NULL) {
			// handle modifiers separately because there are so few of them
			switch (at_scancode) {
				case 0x12:
					// left shift
					*modifier = true;
					*hid_scancode = HID_KEYBOARD_MODIFER_LEFTSHIFT;
					break;

				case 0x59:
					// right shift
					*modifier = true;
					*hid_scancode = HID_KEYBOARD_MODIFER_RIGHTSHIFT;
					break;

				case 0x14:
					if (got_2byte)
						// right ctrl
						*hid_scancode = HID_KEYBOARD_MODIFER_RIGHTCTRL;
					else
						// left ctrl
						*hid_scancode = HID_KEYBOARD_MODIFER_LEFTCTRL;
					*modifier = true;
					break;

				case 0x11:
					if (got_2byte)
						// right alt
						*hid_scancode = HID_KEYBOARD_MODIFER_RIGHTALT;
					else
						// left alt
						*hid_scancode = HID_KEYBOARD_MODIFER_LEFTALT;
					*modifier = true;
					break;

				default:
					*modifier = false;
					*hid_scancode = table[at_scancode];
			}
		}

		*pressed = !got_keyup;
		// reset status flags
		got_keyup = false;
		got_2byte = false;
		got_4byte = false;
		got_8byte = false;
		return (*hid_scancode != 0x00);	// return false if it's an invalid scancode
	}
	return false;

error:
	// reset status and exit, ignoring keycode
	got_keyup = false;
	got_2byte = false;
	got_4byte = false;
	got_8byte = false;
	return false;
}
